#include <cmath>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <unistd.h>
#include <fcntl.h>
#include <vector>

using namespace std;

#define MAXITER 20
#define BEAMSPEED 0.1
#define PRECISION 0.001
#define PROJDIST 0.7
#define EYEDIST  7
#define MAXTRACE ((EYEDIST+10)/BEAMSPEED)
#define MAXREFLECTSOURCE 0

static const unsigned XDIM = 512/3;
static const unsigned YDIM = 384/3;

static int ALKUY=0;
#include <linux/fb.h>
static class Video
{
    int fd;
    unsigned char *p;
    unsigned xres;
public:
    Video() : xres(1504)
    {
        fd = open("/dev/fb0", O_RDWR);
        
        struct fb_cmap cmap;
        cmap.start=0;
        cmap.len  =256;
        cmap.red  =(__u16 *)calloc(256, 2);
        cmap.green=(__u16 *)calloc(256, 2);
        cmap.blue =(__u16 *)calloc(256, 2);
        cmap.transp=NULL;

        ioctl(fd, FBIOGETCMAP, &cmap);
        /* This palette is RGB and used by XAWTV ;) */
        const int r=5;
        const int g=9;
        const int b=5;
        const int rs = 256 / (r - 1);
        const int gs = 256 / (g - 1);
        const int bs = 256 / (b - 1);
        for (int i = 0; i < r*g*b; i++)
        {
            cmap.green[i+16] = (gs * ((i / (r * b)) % g)) * 255;
            cmap.red[i+16]   = (rs * ((i / b) % r)) * 255;
            cmap.blue[i+16]  = (bs * ((i) % b)) * 255;
        }
        ioctl(fd, FBIOPUTCMAP, &cmap);

        free(cmap.red);
        free(cmap.green);
        free(cmap.blue);

        p = (unsigned char *)mmap(NULL, xres*1024, PROT_WRITE, MAP_SHARED, fd, 0);
        if((int)p==-1)
        {
            perror("mmap");
            throw;
        }
    }
    ~Video()
    {
        munmap(p, xres*1024);
    }
    void operator() (unsigned x, unsigned y, unsigned char c)
    {
        /*p[y*xres+x] = c;*/
        
        x+=(1280-XDIM*2)/4; y+=ALKUY;
        
        p[(y*2  )*xres+(x*2  )] = c;
        p[(y*2  )*xres+(x*2+1)] = c;
        p[(y*2+1)*xres+(x*2  )] = c;
        p[(y*2+1)*xres+(x*2+1)] = c;
    }
} PSet;


class Matrix
{
    double val[4][4];
public:
    Matrix()
    {
    }
    Matrix(double ax, double ay, double az)
    {
        InitRotate(ax, ay, az);
    }
    void InitRotate(double ax, double ay, double az)
    {
        double tmp[3][4][4];

        memset(&tmp, 0, sizeof(tmp));

        tmp[0][1][1]=tmp[0][2][2]=cos(ax); tmp[0][2][1]=-(tmp[0][1][2]=sin(ax));
        tmp[1][0][0]=tmp[1][2][2]=cos(ay); tmp[1][0][2]=-(tmp[1][2][0]=sin(ay));
        tmp[2][0][0]=tmp[2][1][1]=cos(az); tmp[2][1][0]=-(tmp[2][0][1]=sin(az));

        tmp[0][0][0] = 1;
        tmp[0][3][3] = 1;
        tmp[1][1][1] = 1;
        tmp[1][3][3] = 1;
        tmp[2][2][2] = 1;
        tmp[2][3][3] = 1;

        for(int i=0; i<4; i++)
        {
            double m2[4];

            for(int j=0; j<4; j++)
                m2[j] = tmp[1][i][0] * tmp[0][0][j] +
                        tmp[1][i][1] * tmp[0][1][j] +
                        tmp[1][i][2] * tmp[0][2][j] +
                        tmp[1][i][3] * tmp[0][3][j];

            for(int j=0; j<4; j++)
                val[i][j] = m2[0] * tmp[2][0][j] +
                            m2[1] * tmp[2][1][j] +
                            m2[2] * tmp[2][2][j] +
                            m2[3] * tmp[2][3][j];
        }
    }
    template<class T>
    void Rotate(T &p)
    {
        register double tx = p.x*val[0][0] + p.y*val[1][0] + p.z*val[2][0] + val[3][0];
        register double ty = p.x*val[0][1] + p.y*val[1][1] + p.z*val[2][1] + val[3][1];

        p.z = p.x*val[0][2] + p.y*val[1][2] + p.z*val[2][2] + val[3][2];
        p.y = ty;
        p.x = tx;
    }
};

class XYZ
{
public:
    double x,y,z;
    
    XYZ() : x(0),y(0),z(0)
    {
    }
    
    XYZ(double xx,double yy,double zz) : x(xx),y(yy),z(zz)
    {
    }
    
    const XYZ &operator +=(const XYZ &p) { x+=p.x; y+=p.y; z+=p.z; return *this; }
    const XYZ &operator -=(const XYZ &p) { x-=p.x; y-=p.y; z-=p.z; return *this; }
    const XYZ &operator /=(const double &p) { x/=p; y/=p; z/=p; return *this; }
    const XYZ &operator *=(const double &p) { x*=p; y*=p; z*=p; return *this; }
    const XYZ operator +(const XYZ &p) const { return XYZ(x+p.x, y+p.y, z+p.z); }
    const XYZ operator -(const XYZ &p) const { return XYZ(x-p.x, y-p.y, z-p.z); }
    const XYZ operator /(const double &p) const { return XYZ(x/p, y/p, z/p); }
    const XYZ operator *(const double &p) const { return XYZ(x*p, y*p, z*p); }
    
    const double dist(const XYZ &p)const 
    {
        XYZ d(x-p.x, y-p.y, z-p.z);
        return sqrt(d.x*d.x + d.y*d.y + d.z*d.z);
    }
    const double length() const
    {
        return sqrt(x*x + y*y + z*z);
    }
    
    inline void Rotate(Matrix &m)
    {
        m.Rotate(*this);
    }
};

class RGB
{
    double r,g,b;
public:
    RGB(double R, double G, double B) : r(R), g(G), b(B)
    {
    }
    const RGB &operator *=(const RGB &p) { r*=p.r; g*=p.g; b*=p.b; return *this; }
    const RGB &operator +=(const RGB &p) { r+=p.r; g+=p.g; b+=p.b; return *this; }
    const RGB &operator -=(const RGB &p) { r-=p.r; g-=p.g; b-=p.b; return *this; }
    const RGB &operator *=(const double &p) { r*=p; g*=p; b*=b; return *this; }
    const RGB &operator /=(const double &p) { r/=p; g/=p; b/=b; return *this; }
    const RGB operator *(const RGB &p) { return RGB(r*p.r, g*p.g, b*p.b); }
    const RGB operator +(const RGB &p) { return RGB(r+p.r, g+p.g, b+p.b); }
    const RGB operator -(const RGB &p) { return RGB(r-p.r, g-p.g, b-p.b); }
    const RGB operator +(const double &p) { return RGB(r+p, g+p, b+p); }
    const RGB operator -(const double &p) { return RGB(r-p, g-p, b-p); }
    const RGB operator *(const double &p) { return RGB(r*p, g*p, b*p); }
    const RGB operator /(const double &p) { return RGB(r/p, g/p, b/p); }

    const unsigned char getR() const { return int(255 * (r<0 ? 0 : r>1 ? 1 : r)); }
    const unsigned char getG() const { return int(255 * (g<0 ? 0 : g>1 ? 1 : g)); }
    const unsigned char getB() const { return int(255 * (b<0 ? 0 : b>1 ? 1 : b)); }
    
    const double power() const { return r+g+b; }
    
    void debug()
    {
//      printf("Returning %.2f %.2f %.2f\n", r,g,b);
    }
};

static class Fract
{
	bool mandelmap[YDIM][XDIM];
    inline bool ismandel(double RP,double IP)
    {
        if(IP<-1.15 || IP>1.15
        || RP<-2.02 || RP>0.5)return false;
        register double r = RP, i = IP;
        for(register int c=0; ; )
        {
            register const double r2 = r*r;
            register const double i2 = i*i;
            
            if(r2+i2 >= 4)
            {
                /* Would never return back to <4. Not part of set. Exit. */
                return false;
            }
            
            i = r*i*2 + IP;
            r = r2-i2 + RP;
            
            if(++c >= MAXITER)
            {
                /* Tried enough. We are inside. */
                return true;
            }
        }
    }

    inline bool isjulia(double XP,double YP, double RP,double IP)
    {
        /* Mandelbrot-rajat */
        if(IP<-1.15 || IP>1.15
        || RP<-2.02 || RP>0.5)return false;

        if(!mandelmap[unsigned((IP+1.15)*YDIM/(1.15*2))]
                     [unsigned((RP+2.02)*XDIM/(2.52))])return false;
        
        /* Julia-rajat */
        if(XP<-2 || XP>2
        || YP<-2 || YP>2)return false;
        
        register double r = XP, i = YP;
        for(register int c=0; ; )
        {
            register const double r2 = r*r;
            register const double i2 = i*i;
            
            if(r2+i2 >= 4)
            {
                /* Would never return back to <4. Not part of set. Exit. */
                return false;
            }
            
            i = r*i*2 + IP;
            r = r2-i2 + RP;
            
            if(++c >= MAXITER)
            {
                /* Tried enough. We are inside. */
                return true;
            }
        }
    }
    
public:
    Fract()
    {
    	double ip=-1.15, iadd=(1.15*2) / YDIM;
    	double rp, radd=2.52/XDIM;
    	for(unsigned y=0; y<YDIM; ++y)
    	{
    		rp = -2.02;
    		for(unsigned x=0; x<XDIM; ++x)
    		{
    			mandelmap[y][x] = ismandel(rp, ip);
    			rp += radd;
    		}
    		ip += iadd;
    	}
    }
    
    inline bool hit(double x,double y, double r,double i)
    {
        return isjulia(x,y,r,i);
    }
} Fract;

static const double randang()
{
    return drand48() * 2 * M_PI;
}

class Light
{
public:
    XYZ pos;
    RGB type;
    double rad;
    
    Light(const XYZ &p, const RGB &c, const double &r) : pos(p), type(c), rad(r)
    {
    }
};

static vector<Light> Lights;

enum tracetypes
{
    find_solid,
    find_light,
    find_reflection
};

static const double w = -0.1; /* the time axis position */
static XYZ endpoint;

static RGB RayTrace(XYZ beam, XYZ dir, tracetypes tracetype)
{
    for(unsigned step=0; ; ++step)
    {
        for(unsigned l=0; l<Lights.size(); ++l)
        {
        	const double rad = Lights[l].rad;
        	double dist = beam.dist(Lights[l].pos);
            if(dist < rad)
            {
            	/* Hit the light source. */
                return Lights[l].type;
            }
        }
        
        if(Fract.hit(beam.z, w, beam.x, beam.y))
        {
        	/* Find the most exact position of hit */
        	double prec = BEAMSPEED;
        	XYZ test = dir;
        	if(prec > PRECISION)
        	{
        		beam -= test;
        		for(;;)
        		{
            		/* beamin osoittamasta kohdasta ei lydy mitn */
            		test /= 2;
            		prec /= 2;
            		beam += test;
            		if(Fract.hit(beam.z, w, beam.x, beam.y))
            		{
            			/* Osui. Tarpeeksi tarkkuutta? */
            			if(prec <= PRECISION)break;
            			
            			/* Ei, palataan viel taakse */
            			beam -= test;
            		}
        		}
        	}
        
        	/* Hit a solid surface. */
        	/* See what we are doing. */
        	
            static const RGB pigment(1, 0.8, 0.5);
            
            RGB result(0, 0, 0);
            if(tracetype != find_light)
            {
            	/* We want to know how well is this surface lighted. */
                for(unsigned l=0; l<Lights.size(); ++l)
                {
                    dir = Lights[l].pos - beam;
                    dir /= dir.length();
                    dir *= BEAMSPEED;
           
                    result += RayTrace(beam+dir, dir, find_light);
                }
            }
            if(tracetype == find_solid)
            {
            	/* Possible reflections affect also. */
                for(unsigned l=0; l<MAXREFLECTSOURCE; ++l)
                {
                    /* Random direction */
                    double ang = randang();
                    dir.x = cos(ang);
                    dir.y = sin(ang);
                    ang = randang();
                    dir.x *= sin(ang);
                    dir.z = cos(ang);
                    dir *= BEAMSPEED;
                    
                    result += RayTrace(beam+dir, dir, find_reflection) * pigment;
                }
            }
            result *= pigment;
            
            result.debug();
            return result;
        }

        if(step >= MAXTRACE)
        {
        	endpoint = beam;
            /* Found *nothing* in this direction */
            return RGB(0, 0, 0);
        }
        
        beam += dir;
    }
}

int main(int argc, const char *const *argv)
{
    Lights.push_back(Light(XYZ(2.4,    0,    0),   RGB(0.30, 1.00, 0.40), 0.2));   /* Vihre valo */
    Lights.push_back(Light(XYZ(0,   -2.5,  2.5),   RGB(1.00, 0.60, 0.10), 0.4));   /* Punainen valo */
    Lights.push_back(Light(XYZ(-1.0, 2.5, -1.3),   RGB(0.10, 0.30, 1.00), 0.6));   /* Sininen valo */
    
    int Parilliset = 1;
    int Parittomat = 1;
    
    for(int a=1; a<argc; ++a)
    {
    	const char *s = argv[a];
    	if(!strcmp(s, "2"))Parittomat = 0;
    	if(!strcmp(s, "1"))Parilliset = 0;
    }
/*
    for(unsigned x=0; x<256; ++x)
        for(unsigned y=0; y<260; ++y)
            PSet(x, y, x);
    return;
*/

	if(Parilliset && !Parittomat) ALKUY = YDIM;
	else if(Parittomat && !Parilliset) ALKUY = 0;
	else ALKUY = (1024-YDIM*2)/4;

    for(double ax=0; ax<M_PI*4; ax+=M_PI / 100)
    {
        Matrix eyematrix(ax-M_PI, sin(ax), -0.3 - sin(ax/2));
        
        char fnBuf[64]; static unsigned imgind=1000;
        unsigned char tgaBuf[18] = {0,0,2,0,0,0,0, 0,0,0,0,0, XDIM&255, XDIM>>8, YDIM&255, YDIM>>8, 24, 32};
        sprintf(fnBuf, "tga/mand%04u.tga", ++imgind);
        
        if((imgind & 1)==1 && !Parittomat)continue;
        if((imgind & 1)==0 && !Parilliset)continue;
        
        FILE *fp = fopen(fnBuf, "wb");
        fwrite(tgaBuf, 1, 18, fp);

        XYZ beam(0,0, -EYEDIST);
        beam.Rotate(eyematrix);

        for(unsigned y=0; y<YDIM; ++y)
        {
            const double ydisp = (y / (double)(YDIM-1)*2) - 1;
            
            for(unsigned x=0; x<XDIM; ++x)
            {
                const double xdisp = (x / (double)(XDIM-1)*2) - 1.5;
                
                XYZ proj(xdisp,ydisp,EYEDIST-PROJDIST);
                XYZ dir = proj * (BEAMSPEED / proj.length()); dir.Rotate(eyematrix);
                
                #if 0
                
                XYZ proj1(proj.x,            proj.y - 0.6/YDIM, proj.z);
                XYZ proj2(proj.x - 0.4/XDIM, proj.y + 0.3/YDIM, proj.z);
                XYZ proj3(proj.x + 0.4/XDIM, proj.y + 0.3/YDIM, proj.z);
                
                XYZ dir1 = proj1 * (BEAMSPEED / proj1.length()); dir1.Rotate(eyematrix);
                XYZ dir2 = proj2 * (BEAMSPEED / proj2.length()); dir2.Rotate(eyematrix);
                XYZ dir3 = proj3 * (BEAMSPEED / proj3.length()); dir3.Rotate(eyematrix);
                #endif
                
                PSet(x,y, 15);
                RGB pix = RayTrace(beam, dir, find_solid); 
                #if 0
                RGB pix1 = RayTrace(beam, dir1, find_solid); 
                RGB pix2 = RayTrace(beam, dir2, find_solid);
                RGB pix3 = RayTrace(beam, dir3, find_solid);
                #endif
                
                unsigned char R = pix.getR();
                unsigned char G = pix.getG();
                unsigned char B = pix.getB();
                
                if(R==0 && G==0 && B==0)
                {
                    for(unsigned l=0; l<Lights.size(); ++l)
                    {
                    	double dist = endpoint.dist(Lights[l].pos);
                    	if(dist < 8)
                    	{
	                    	pix += Lights[l].type * (8-dist);
	                    }
                    }
                	
                	R = pix.getR();
                	G = pix.getG();
                	B = pix.getB();
                }

                putc(B, fp);
                putc(G, fp);
                putc(B, fp);
                
                unsigned char c = (B*6/256)
                                + (R*6/256) * 5
                                + (G*10/256) * (5*5);
                PSet(x,y, 16 + c);
            }
        }
        
        fclose(fp);
    }
    return 0;
}
