#include <cmath>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <unistd.h>
#include <fcntl.h>
#include <vector>

using namespace std;

static const unsigned XDIM = 512;
static const unsigned YDIM = XDIM*7/13;

#define INITIAL_W -0.1
#define USEMANDMAP 0
#define DRAWDOUBLEPIX 1
#define DISPLAYRENDER 1

#define MAXITER 11
#define BEAMSPEED 0.01
#define PRECISION 0.0002
#define PROJSIZEX 1
#define PROJSIZEY (XDIM/YDIM)
#define PROJDIST 1.6
#define EYEDIST  3.6
#define MAXTRACE ((EYEDIST+2)/BEAMSPEED)
#define MAXREFLECTSOURCE 0

#define TRIANGLETOP 0.10825317546076747171    //  1/8 * sin(60)
#define TRIANGLESIDES 0.0625                  //  1/8 * 0.5

static int ALKUY=0;
#include <linux/fb.h>
static class Video
{
#if DISPLAYRENDER
    int fd;
    unsigned char *p;
    unsigned xres;
public:
    Video() : xres(1504)
    {
        fd = open("/dev/fb0", O_RDWR);
        
        struct fb_cmap cmap;
        cmap.start=0;
        cmap.len  =256;
        cmap.red  =(__u16 *)calloc(256, 2);
        cmap.green=(__u16 *)calloc(256, 2);
        cmap.blue =(__u16 *)calloc(256, 2);
        cmap.transp=NULL;

        ioctl(fd, FBIOGETCMAP, &cmap);
        /* This palette is RGB and used by XAWTV ;) */
        const int r=5;
        const int g=9;
        const int b=5;
        const int rs = 256 / (r - 1);
        const int gs = 256 / (g - 1);
        const int bs = 256 / (b - 1);
        for (int i = 0; i < r*g*b; i++)
        {
            cmap.green[i+16] = (gs * ((i / (r * b)) % g)) * 255;
            cmap.red[i+16]   = (rs * ((i / b) % r)) * 255;
            cmap.blue[i+16]  = (bs * ((i) % b)) * 255;
        }
        ioctl(fd, FBIOPUTCMAP, &cmap);

        free(cmap.red);
        free(cmap.green);
        free(cmap.blue);

        p = (unsigned char *)mmap(NULL, xres*1024, PROT_WRITE, MAP_SHARED, fd, 0);
        if((int)p==-1)
        {
            perror("mmap");
            throw;
        }
    }
    ~Video()
    {
        munmap(p, xres*1024);
    }
    void operator() (unsigned x, unsigned y, unsigned char c)
    {
        y += ALKUY;
        #if DRAWDOUBLEPIX
        if(XDIM<1280/2)x+=(1280-XDIM*2)/4;
        p[(y*2  )*xres+(x*2  )] = c;
        p[(y*2  )*xres+(x*2+1)] = c;
        p[(y*2+1)*xres+(x*2  )] = c;
        p[(y*2+1)*xres+(x*2+1)] = c;
        #else
        p[y*xres+x] = c;
        #endif
    }
#else
public:
    inline void operator() (unsigned, unsigned, unsigned char)
    {
    }
#endif
} PSet;


class Matrix
{
    double val[4][4];
public:
    Matrix()
    {
    }
    Matrix(double ax, double ay, double az)
    {
        InitRotate(ax, ay, az);
    }
    void InitRotate(double ax, double ay, double az)
    {
        double tmp[3][4][4];

        memset(&tmp, 0, sizeof(tmp));

        tmp[0][1][1]=tmp[0][2][2]=cos(ax); tmp[0][2][1]=-(tmp[0][1][2]=sin(ax));
        tmp[1][0][0]=tmp[1][2][2]=cos(ay); tmp[1][0][2]=-(tmp[1][2][0]=sin(ay));
        tmp[2][0][0]=tmp[2][1][1]=cos(az); tmp[2][1][0]=-(tmp[2][0][1]=sin(az));

        tmp[0][0][0] = 1;
        tmp[0][3][3] = 1;
        tmp[1][1][1] = 1;
        tmp[1][3][3] = 1;
        tmp[2][2][2] = 1;
        tmp[2][3][3] = 1;

        for(int i=0; i<4; i++)
        {
            double m2[4];

            for(int j=0; j<4; j++)
                m2[j] = tmp[1][i][0] * tmp[0][0][j] +
                        tmp[1][i][1] * tmp[0][1][j] +
                        tmp[1][i][2] * tmp[0][2][j] +
                        tmp[1][i][3] * tmp[0][3][j];

            for(int j=0; j<4; j++)
                val[i][j] = m2[0] * tmp[2][0][j] +
                            m2[1] * tmp[2][1][j] +
                            m2[2] * tmp[2][2][j] +
                            m2[3] * tmp[2][3][j];
        }
    }
    template<class T>
    void Rotate(T &p)
    {
        register double tx = p.x*val[0][0] + p.y*val[1][0] + p.z*val[2][0] + val[3][0];
        register double ty = p.x*val[0][1] + p.y*val[1][1] + p.z*val[2][1] + val[3][1];

        p.z = p.x*val[0][2] + p.y*val[1][2] + p.z*val[2][2] + val[3][2];
        p.y = ty;
        p.x = tx;
    }
};

class XYZ
{
public:
    double x,y,z;
    bool nowhere;
    
    XYZ() : x(0),y(0),z(0), nowhere(true)
    {
    }
    XYZ(const XYZ &p) : x(p.x), y(p.y), z(p.z), nowhere(p.nowhere)
    {
    }
    
    XYZ(double xx,double yy,double zz) : x(xx),y(yy),z(zz), nowhere(false)
    {
    }
    
    const XYZ &operator +=(const XYZ &p) { x+=p.x; y+=p.y; z+=p.z; return *this; }
    const XYZ &operator -=(const XYZ &p) { x-=p.x; y-=p.y; z-=p.z; return *this; }
    const XYZ &operator /=(const double &p) { x/=p; y/=p; z/=p; return *this; }
    const XYZ &operator *=(const double &p) { x*=p; y*=p; z*=p; return *this; }
    const XYZ operator +(const XYZ &p) const { return XYZ(x+p.x, y+p.y, z+p.z); }
    const XYZ operator -(const XYZ &p) const { return XYZ(x-p.x, y-p.y, z-p.z); }
    const XYZ operator /(const double &p) const { return XYZ(x/p, y/p, z/p); }
    const XYZ operator *(const double &p) const { return XYZ(x*p, y*p, z*p); }
    const XYZ &operator= (const XYZ &p)
    {
        x = p.x;
        y = p.y;
        z = p.z;
        nowhere = p.z;
        return *this;
    }
    
    const double dist(const XYZ &p)const 
    {
        XYZ d(x-p.x, y-p.y, z-p.z);
        return sqrt(d.x*d.x + d.y*d.y + d.z*d.z);
    }
    const double length() const
    {
        return sqrt(x*x + y*y + z*z);
    }
    
    inline void Rotate(Matrix &m)
    {
        m.Rotate(*this);
    }
};

class RGB
{
    double r,g,b;
    
    inline const int intensity(double elem) const
    {
        if(elem < 0)return 0;
        const double maxi = max(max(r,g), b);
        if(maxi <= 1)return int(255*elem);
        if(elem > 1)return 255;
        const double jaettavaa = maxi-1;
        int v = int((elem + jaettavaa)*255);
        return v>=255 ? 255 : v;
    }
    
public:
	RGB() : r(0), g(0), b(0)
	{
	}
    RGB(double R, double G, double B) : r(R), g(G), b(B)
    {
    }
    const RGB &operator *=(const RGB &p) { r*=p.r; g*=p.g; b*=p.b; return *this; }
    const RGB &operator +=(const RGB &p) { r+=p.r; g+=p.g; b+=p.b; return *this; }
    const RGB &operator -=(const RGB &p) { r-=p.r; g-=p.g; b-=p.b; return *this; }
    const RGB &operator *=(const double &p) { r*=p; g*=p; b*=b; return *this; }
    const RGB &operator /=(const double &p) { r/=p; g/=p; b/=b; return *this; }
    const RGB operator *(const RGB &p) { return RGB(r*p.r, g*p.g, b*p.b); }
    const RGB operator +(const RGB &p) { return RGB(r+p.r, g+p.g, b+p.b); }
    const RGB operator -(const RGB &p) { return RGB(r-p.r, g-p.g, b-p.b); }
    const RGB operator +(const double &p) { return RGB(r+p, g+p, b+p); }
    const RGB operator -(const double &p) { return RGB(r-p, g-p, b-p); }
    const RGB operator *(const double &p) { return RGB(r*p, g*p, b*p); }
    const RGB operator /(const double &p) { return RGB(r/p, g/p, b/p); }
    
    const unsigned char getR() const { return intensity(r); }
    const unsigned char getG() const { return intensity(g); }
    const unsigned char getB() const { return intensity(b); }
    
    const double power() const { return r+g+b; }
    
    void debug()
    {
//      printf("Returning %.2f %.2f %.2f\n", r,g,b);
    }
};

#define FLOAT_ARG double   //const double &

static class Fract
{
public:
    static const double xmin = -2.0;
    static const double xmax = +2.0;
    static const double ymin = -1.3;
    static const double ymax = +1.3;
    static const double rmin = -2.02;
    static const double rmax = +0.5;
    static const double imin = -1.15;
    static const double imax = +1.15;
private:
#if USEMANDMAP
    bool mandelmap[YDIM][XDIM];
#endif
    inline bool ismandel(FLOAT_ARG RP, FLOAT_ARG IP)
    {
        if(IP<imin || IP>imax
        || RP<rmin || RP>rmax)return false;
        register double r = RP, i = IP;
        for(register int c=0; ; )
        {
            register const double r2 = r*r;
            register const double i2 = i*i;
            
            if(r2+i2 >= 4)
            {
                /* Would never return back to <4. Not part of set. Exit. */
                return false;
            }
            
            i = r*i*2 + IP;
            r = r2-i2 + RP;
            
            if(++c >= MAXITER)
            {
                /* Tried enough. We are inside. */
                return true;
            }
        }
    }

    inline bool isjulia(FLOAT_ARG XP, FLOAT_ARG YP, FLOAT_ARG RP, FLOAT_ARG IP)
    {
        if(wayout(XP,YP, RP,IP))return false;
#if USEMANDMAP
        if(!mandelmap[unsigned((IP-imin)*YDIM/(imax-imin))]
                     [unsigned((RP-rmin)*XDIM/(rmax-rmin))])return false;
#endif
        register double r = XP, i = YP;
        for(register int c=0; ; )
        {
            register const double r2 = r*r;
            register const double i2 = i*i;
            
            if(r2+i2 >= 4)
            {
                /* Would never return back to <4. Not part of set. Exit. */
                return false;
            }
            
            i = r*i*2 + IP;
            r = r2-i2 + RP;
            
            if(++c >= MAXITER)
            {
                /* Tried enough. We are inside. */
                return true;
            }
        }
    }
    
public:
    Fract()
    {
#if USEMANDMAP
        double ip = imin;
        const double iadd=(imax-imin) / YDIM;
        const double radd=(rmax-rmin) / XDIM;
        for(unsigned y=0; y<YDIM; ++y)
        {
            double rp = rmin;
            for(unsigned x=0; x<XDIM; ++x)
            {
                mandelmap[y][x] = ismandel(rp, ip);
                rp += radd;
            }
            ip += iadd;
        }
#endif
    }
    
    /* Paluuarvot: true=sisll, false=ulkona */
    inline bool hit(FLOAT_ARG x, FLOAT_ARG y, FLOAT_ARG r, FLOAT_ARG i)
    {
        return isjulia(x,y,r,i);
    }
    
    /* Paluuarvot: true=tosi ulkona, false=ehk sisll */
    bool wayout(FLOAT_ARG x, FLOAT_ARG y, FLOAT_ARG r, FLOAT_ARG i)
    {
        /* Mandelbrot-rajat */
        if(i<imin || i>imax
        || r<rmin || r>rmax)return true;

        /* Julia-rajat */
        if(x<xmin || x>xmax
        || y<ymin || y>ymax)return true;
        
        return false;
    }
} Fract;

static const double randang()
{
    return drand48() * 2 * M_PI;
}

class Light
{
public:
    XYZ pos, oldpos;
    RGB type;
    double rad;
    double eff;
    
    Light(const XYZ &p, const RGB &c, const double &r) : pos(p), type(c), rad(r), eff(1)
    {
        oldpos = pos;
    }
};

static vector<Light> Lights;

static const double w = INITIAL_W; /* the time axis position */
static XYZ endpoint;

/* return value: false=mahiksia, true=ulos menossa */
static bool SkipEmpty(XYZ &beam, const XYZ &dir, unsigned &step)
{
/*
    printf("z %.6f += %.6f\n"
           "x %.6f += %.6f\n"
           "y %.6f += %.6f\n",
           beam.z, dir.z,
           beam.x, dir.x,
           beam.y, dir.y);
*/
    if(((beam.z < Fract.xmin && dir.z <= 0) || (beam.z > Fract.xmax && dir.z >= 0))
    || ((beam.x < Fract.rmin && dir.x <= 0) || (beam.x > Fract.rmax && dir.x >= 0))
    || ((beam.y < Fract.imin && dir.y <= 0) || (beam.y > Fract.imax && dir.y >= 0))
    || ((w      < Fract.ymin              ) || (w      > Fract.ymax              ))
    )
    {
        double minimi = MAXTRACE*BEAMSPEED;
        #if 0
        for(unsigned l=0; l<Lights.size(); ++l)
        {
            double dist = beam.dist(Lights[l].pos);
            if(dist < minimi)minimi = dist;
        }
        #endif
        if(minimi >= BEAMSPEED)
        {
            double add = minimi / dir.length();
            step += int(add);
            beam += dir * add;
        }

        /* Ulos menossa. */
        return true;
    }
    
    #if 0
    
    double zspace = 0;
    double wspace = 0;
    double xspace = 0;
    double yspace = 0;

    if(beam.z < Fract.xmin && dir.z > 0)zspace = Fract.xmin - beam.z; else if(beam.z > Fract.xmax && dir.z < 0)zspace = beam.z - Fract.xmax;
    if(beam.x < Fract.rmin && dir.x > 0)xspace = Fract.rmin - beam.x; else if(beam.x > Fract.rmax && dir.x < 0)xspace = beam.x - Fract.rmax;
    if(beam.y < Fract.imin && dir.y > 0)yspace = Fract.imin - beam.y; else if(beam.y > Fract.imax && dir.y < 0)yspace = beam.y - Fract.imax;
    if(w      < Fract.ymin             )wspace = Fract.ymin - w     ; else if(w      > Fract.ymax             )wspace = w      - Fract.ymax;
    
    //double minimi = sqrt(zspace*zspace + wspace*wspace + xspace*xspace + yspace*yspace) - BEAMSPEED;
    double minimi = max(max(zspace,wspace), max(xspace,yspace)) - BEAMSPEED;
    
    #if 0
    for(unsigned l=0; l<Lights.size(); ++l)
    {
        double dist = beam.dist(Lights[l].pos);
        if(dist < minimi)minimi = dist;
    }
    #endif
    if(minimi >= BEAMSPEED)
    {
        double add = minimi / dir.length();
        step += int(add);
        beam += dir * add;
    }
    #endif

    return false;
}

enum tracetypes
{
    find_solid,
    find_light,
    find_reflection
};

static XYZ FindEnd(XYZ beam, const XYZ &dir)
{
	while(Fract.hit(beam.z, w, beam.x, beam.y))
	{
		beam -= dir;
	}
    unsigned step=0;
    bool ulos = SkipEmpty(beam, dir, step);
    if(ulos)
    {
        beam.nowhere = true;
        beam += dir * MAXTRACE;
        return beam;
    }
    
    for(; ; ++step)
    {
        if(Fract.hit(beam.z, w, beam.x, beam.y))
        {
            /* Find the most exact position of hit */
            double prec = BEAMSPEED;
            XYZ test = dir;
            if(prec > PRECISION)
            {
                beam -= test;
                for(;;)
                {
                    /* beamin osoittamasta kohdasta ei lydy mitn */
                    test /= 2;
                    prec /= 2;
                    beam += test;
                    if(Fract.hit(beam.z, w, beam.x, beam.y))
                    {
                        /* Osui. Tarpeeksi tarkkuutta? */
                        if(prec <= PRECISION)break;
                        
                        /* Ei, palataan viel taakse */
                        beam -= test;
                    }
                }
            }
            return beam;
        }
        if(step >= MAXTRACE)
        {
            /* This direction didn't hit at all */
            beam.nowhere = true;
            return beam;
        }
        beam += dir;
    }
}

static RGB RayTrace(XYZ beam, const XYZ &dir, tracetypes tracetype)
{
    endpoint.nowhere = false;
    
    unsigned step=0;    
    bool ulos = SkipEmpty(beam, dir, step);
    if(ulos)
    {
        if(tracetype != find_light)
        {
            endpoint = beam;
            endpoint.nowhere = true;
            /* Found *nothing* in this direction */
            return RGB(0, 0, 0);
        }
        /* Would never hit anything */
        for(; ; ++step)
        {
            for(unsigned l=0; l<Lights.size(); ++l)
            {
                const double rad = Lights[l].rad;
                double dist = beam.dist(Lights[l].pos);
                if(dist < rad)
                {
                    /* Hit the light source. */
                    return Lights[l].type;
                }
            }
            if(step >= MAXTRACE)
            {
                endpoint = beam;
                endpoint.nowhere = true;
                /* Found *nothing* in this direction */
                return RGB(0, 0, 0);
            }
            
            beam += dir;
        }
    }
    
    for(; ; ++step)
    {
        if(tracetype == find_light)
            for(unsigned l=0; l<Lights.size(); ++l)
            {
                const double rad = Lights[l].rad;
                double dist = beam.dist(Lights[l].pos);
                if(dist < rad)
                {
                    /* Hit the light source. */
                    return Lights[l].type;
                }
            }
        
        if(Fract.hit(beam.z, w, beam.x, beam.y))
        {
            /* Find the most exact position of hit */
            double prec = BEAMSPEED;
            XYZ test = dir;
            if(prec > PRECISION)
            {
                beam -= test;
                for(;;)
                {
                    /* beamin osoittamasta kohdasta ei lydy mitn */
                    test /= 2;
                    prec /= 2;
                    beam += test;
                    if(Fract.hit(beam.z, w, beam.x, beam.y))
                    {
                        /* Osui. Tarpeeksi tarkkuutta? */
                        if(prec <= PRECISION)break;
                        
                        /* Ei, palataan viel taakse */
                        beam -= test;
                    }
                }
            }
        
            /* Hit a solid surface. */
            /* See what we are doing. */
            
            static const RGB pigment(0.8, 0.8, 0.8);
            
            RGB result(0, 0, 0);
            if(tracetype != find_light)
            {
                /* We want to know how well is this surface lighted. */
                for(unsigned l=0; l<Lights.size(); ++l)
                {
                    XYZ tmp(Lights[l].pos - beam);
                    tmp *= BEAMSPEED / tmp.length();
                    result += RayTrace(beam+tmp, tmp, find_light) * Lights[l].eff;
                }
            }
            if(tracetype == find_solid)
            {
                /* Possible reflections affect also. */
                for(unsigned l=0; l<MAXREFLECTSOURCE; ++l)
                {
                    /* Random direction */
                    double ang = randang(), ang2 = randang();
                    XYZ tmp(cos(ang) * sin(ang2), sin(ang), cos(ang2));
                    tmp *= BEAMSPEED;                    
                    result += RayTrace(beam+tmp, tmp, find_reflection) / (MAXREFLECTSOURCE+2.0);
                }
            }
            result *= pigment;
            
            result.debug();
            return result;
        }

        if(step >= MAXTRACE)
        {
            endpoint = beam;
            endpoint.nowhere = true;
            /* Found *nothing* in this direction */
            return RGB(0, 0, 0);
        }
        
        beam += dir;
    }
}

int main(int argc, const char *const *argv)
{
    Lights.push_back(Light(XYZ(+0.0, -3.0, +3.0),   RGB(0.30, 0.80, 0.10), 0.3));   /* Vihre valo    */
    Lights.push_back(Light(XYZ(-3.0, +3.3, -3.0),   RGB(1.00, 0.40, 0.10), 0.2));   /* Punainen valo  */
    Lights.push_back(Light(XYZ(-1.0, -2.3, -1.0),   RGB(0.80, 0.80, 0.80), 0.5));   /* Valkoinen valo */
    Lights.push_back(Light(XYZ(+3.0, +3.3, +1.0),   RGB(0.10, 0.60, 1.00), 0.4));   /* Sininen valo   */
    
    int Parilliset = 1;
    int Parittomat = 1;
    
    for(int a=1; a<argc; ++a)
    {
        const char *s = argv[a];
        if(!strcmp(s, "2"))Parittomat = 0;
        if(!strcmp(s, "1"))Parilliset = 0;
    }

    if(Parilliset && !Parittomat) ALKUY = YDIM;
    else if(Parittomat && !Parilliset) ALKUY = 0;
    else ALKUY = 0;//(1024-YDIM)/4;

    for(double ax=0; ax<M_PI*4; ax+=M_PI / 100)
    //double ax;
    //for(ax=0, w=-1.3; w<=1.3; w+=0.01)
    {
        Matrix eyematrix(ax-M_PI, sin(ax), -0.3 - sin(ax/2));
        //Matrix lightmatrix(ax*2, cos(ax)-0.9, cos(ax));
        Matrix lightmatrix = eyematrix;
        
        static unsigned imgind=1000;
        static const unsigned char tgaBuf[] = {0,0,2, 0,0, 0,0, 0,  0,0, 0,0, XDIM&255, XDIM>>8, YDIM&255, YDIM>>8, 24,32};
        char fnBuf[64];
        sprintf(fnBuf, "tga/mand%04u.tga", ++imgind);
        
        if((imgind & 1)==1 && !Parittomat)continue;
        if((imgind & 1)==0 && !Parilliset)continue;
        
        FILE *fp = fopen(fnBuf, "wb");
        fwrite(tgaBuf, 1, sizeof tgaBuf, fp);

        XYZ beam(0, 0, -EYEDIST);
        beam.Rotate(eyematrix);
        beam.x -= 0.8;
        
        for(unsigned l=0; l<Lights.size(); ++l)
        {
            Lights[l].pos = Lights[l].oldpos;
            Lights[l].pos.Rotate(lightmatrix);
        }

        for(unsigned y=0; y<YDIM; ++y)
        {
            const double ydisp1= (y               / (double)(YDIM-1) - 0.5) * (PROJSIZEY*2.0);
            const double ydisp = ((y+TRIANGLETOP) / (double)(YDIM-1) - 0.5) * (PROJSIZEY*2.0);
            
            for(unsigned x=0; x<XDIM; ++x)
            {
                const double xdisp0= ((x-TRIANGLESIDES) / (double)(XDIM-1) - 0.5) * (PROJSIZEX*2.0);
                const double xdisp1= ((x+TRIANGLESIDES) / (double)(XDIM-1) - 0.5) * (PROJSIZEX*2.0);
                const double xdisp = (x                 / (double)(XDIM-1) - 0.5) * (PROJSIZEX*2.0);
                
                XYZ proj(xdisp,ydisp,PROJDIST);
                XYZ proj1(xdisp,ydisp1,PROJDIST);
                XYZ proj2(xdisp0,ydisp,PROJDIST);
                XYZ proj3(xdisp1,ydisp,PROJDIST);
                
                XYZ dir = proj * (BEAMSPEED / proj.length()); dir.Rotate(eyematrix);
                XYZ dir1 = proj1 * (BEAMSPEED / proj1.length()); dir1.Rotate(eyematrix);
                XYZ dir2 = proj2 * (BEAMSPEED / proj2.length()); dir2.Rotate(eyematrix);
                XYZ dir3 = proj3 * (BEAMSPEED / proj3.length()); dir3.Rotate(eyematrix);
                
                PSet(x,y, 15);
                
                RGB pix;
                
                XYZ end1 = FindEnd(beam, dir1);
                if(end1.nowhere)
                {
                    for(unsigned l=0; l<Lights.size(); ++l)
                    {
                    	Lights[l].eff = 1;
                    	
                        double dist = pow(end1.dist(Lights[l].pos), 3.3);
                        if(dist<0.01)dist=0.01;
                        double power = 2.0 / dist;
                        pix += Lights[l].type * power;
                    }
                }
                else
                {
                    XYZ end2 = FindEnd(beam + dir2*(end1.length()-beam.length()), dir2);
                    XYZ end3 = FindEnd(beam + dir3*(end1.length()-beam.length()), dir3);
                    XYZ middle = (end1+end2+end3)/3;
                    
                    XYZ p21 = end2-end1;
                    XYZ p31 = end3-end1;

                    XYZ normal(p21.y * p31.z - p31.y * p21.z,
                               p31.x * p21.z - p21.x * p31.z,
                               p21.x * p31.y - p31.x * p21.y);
                    if(normal.length())
                    {
                        normal /= normal.length();

                        for(unsigned l=0; l<Lights.size(); ++l)
                        {
                            XYZ live = Lights[l].pos - middle;
                            live /= live.length();
                            
                            double pitu = normal.x * live.x
                                        + normal.y * live.y
                                        + normal.z * live.z;
                            if(pitu < 0)pitu = 0;
                            
                            Lights[l].eff = pitu;
                        }
                    }
                    pix = RayTrace(middle-dir, dir, find_solid);
                }
                
                unsigned char R = pix.getR();
                unsigned char G = pix.getG();
                unsigned char B = pix.getB();
                
                putc(B, fp);
                putc(G, fp);
                putc(R, fp);
                
                unsigned char c = (B*5/256)
                                + (R*5/256) * 5
                                + (G*9/256) * (5*5);
                PSet(x,y, 16 + c);
            }
        }
        
        fclose(fp);
    }
    return 0;
}
